
#pragma once


///
// Find the templates for coordinate generation
//
// Autogenerated, do not edit.
//
//   $SCHRODINGER/run mol_generator.py templates.mae
//
// generated using templates.mae version c17d2be57897d3e83435.
//


#include <vector>

class sketcherMinimizerMolecule;
namespace schrodinger {

///
// Create a new vector of sketcherMinimizerMolecule*. Caller
// owns the sketcherMinimizerMolecule objects.
std::vector<sketcherMinimizerMolecule*> coordgen_templates();

}
